//Nicolas Boutry, EPITA Research and Development Laboratory (LRDE)
//May 2015
//This program makes images digitally well-composed by removing critical configurations.
// The resulting file is saved to "result.pgm".

//Installation:
//Olena library is necessary to compile this source file
//https://www.lrde.epita.fr/wiki/Olena/Download

#include <cassert>
#include <mln/io/pgm/save.hh>
#include <string.h>
#include <mln/io/pgm/load.hh>
//#include <mln/debug/println.hh>
#include <vector>
#include <queue>
#include <iostream>

using namespace mln;
using value::int_u8;

struct Coords{
  int x;
  int y;
};


unsigned ** alloue_image(int sX, int sY)
{
  unsigned ** im_out = (unsigned**)malloc(sizeof(unsigned*) * sX);
  for (int i = 0; i < sX; i++)
    im_out[i] = (unsigned*)malloc(sizeof(unsigned) * sY);

  return im_out;
}  

unsigned ** recopie_image(unsigned ** im_in, int sX, int sY)
{
  unsigned ** im_out = alloue_image(sX,sY);

  for(int x = 0;x < sX;x++)
    for(int y = 0;y < sY;y++)
      im_out[x][y] = im_in[x][y];

  return im_out;
}  

bool CriticalConfiguration(unsigned a, unsigned b, unsigned c, unsigned d){

      unsigned minad,MAXAD,minbc,MAXBC;
      if(a < d){
	minad = a;
	MAXAD = d;
      } else {
	minad = d;
	MAXAD = a;
      }

      if(b < c){
	minbc = b;
	MAXBC = c;
      } else {
	minbc = c;
	MAXBC = b;
      }
  return ((MAXAD < minbc) || (MAXBC < minad));
}

long nbCCs(unsigned ** u, int sX, int sY){

  long number = 0;
  int x,y;
  for(x = 0; x < sX-1; x++){
    for(y = 0; y < sY-1; y++){
      if(CriticalConfiguration(u[x][y],u[x+1][y],u[x][y+1],u[x+1][y+1])){
	number++;
      }    
    }
  }
  return number;
}

void increasingFIFO(unsigned ** u, int sX, int sY)
{

  std::queue<Coords> myqueue;


  long COMPTEURCCSINIT = 0;


  //Initialization of the queue
  int x,y;
  for(x = 0; x < sX-1;x++){
    for(y = 0; y < sY-1;y++){
      unsigned a = u[x][y];
      unsigned b = u[x+1][y];
      unsigned c = u[x][y+1];
      unsigned d = u[x+1][y+1];

      unsigned minad,MAXAD,minbc,MAXBC;
      //      bool aINFd,bINFc;
      if(a < d){
	minad = a;
	MAXAD = d;
	//	aINFd = true;
      } else {
	minad = d;
	MAXAD = a;
	//	aINFd = false;
      }

      if(b < c){
	minbc = b;
	MAXBC = c;
	//	bINFc = true;
      } else {
	minbc = c;
	MAXBC = b;
	//	bINFc = false;
      }
      //we enqueue the critical configurations
      if((MAXAD < minbc) || (MAXBC < minad)){
	Coords CC;
	CC.x = x;
	CC.y = y;
	myqueue.push(CC);
      }
    }
  }
  //Iteration
  while(!myqueue.empty()){

    //correction of the critical configuration
    Coords coords = myqueue.front();
    int x = coords.x;
    int y = coords.y;
    myqueue.pop();

    unsigned a = u[x][y];
    unsigned b = u[x+1][y];
    unsigned c = u[x][y+1];
    unsigned d = u[x+1][y+1];

    unsigned minad,MAXAD,minbc,MAXBC;
    bool aINFd,bINFc;
    if(a < d){
      minad = a;
      MAXAD = d;
      aINFd = true;
    } else {
      minad = d;
      MAXAD = a;
      aINFd = false;
    }

    if(b < c){
      minbc = b;
      MAXBC = c;
      bINFc = true;
    } else {
      minbc = c;
      MAXBC = b;
      bINFc = false;
    }

    Coords correctedpixel;
    bool detectedCC = false;

    //primary type
    if(MAXAD < minbc){
      detectedCC = true;
      if(aINFd){      
	u[x+1][y+1] = minbc;
	correctedpixel.x = x+1;
	correctedpixel.y = y+1;
      }else{
	u[x][y] = minbc;
	correctedpixel.x = x;
	correctedpixel.y = y;
      }
    }

    //secundary type
    if(MAXBC < minad){
      detectedCC = true;
      if(bINFc){
	u[x][y+1] = minad;
	correctedpixel.x = x;
	correctedpixel.y = y+1;
      } else {
	u[x+1][y] = minad;
	correctedpixel.x = x+1;
	correctedpixel.y = y;
      }
    }

    if(detectedCC){
      //detection of new critical configurations due to the possible propagation
      if((correctedpixel.x > 0) && (correctedpixel.y > 0)){
	int newx = correctedpixel.x-1;
	int newy = correctedpixel.y-1;
	bool propagation = CriticalConfiguration(u[newx][newy],u[newx+1][newy],u[newx][newy+1],u[newx+1][newy+1]);
	if(propagation){
	  Coords newCC;
	  newCC.x = newx;
	  newCC.y = newy; 
	  myqueue.push(newCC);
	}
      }
      if((correctedpixel.x < sX-1) && (correctedpixel.y > 0)){
	int newx = correctedpixel.x;
	int newy = correctedpixel.y-1;
	bool propagation = CriticalConfiguration(u[newx][newy],u[newx+1][newy],u[newx][newy+1],u[newx+1][newy+1]);
	if(propagation){
	  Coords newCC;
	  newCC.x = newx;
	  newCC.y = newy; 
	  myqueue.push(newCC);
	}
      }
      if((correctedpixel.x > 0) && (correctedpixel.y < sY-1)){
	int newx = correctedpixel.x-1;
	int newy = correctedpixel.y;
	bool propagation = CriticalConfiguration(u[newx][newy],u[newx+1][newy],u[newx][newy+1],u[newx+1][newy+1]);
	if(propagation){
	  Coords newCC;
	  newCC.x = newx;
	  newCC.y = newy; 
	  myqueue.push(newCC);
	}
      }
      if((correctedpixel.x < sX-1) && (correctedpixel.y < sY-1)){
	int newx = correctedpixel.x;
	int newy = correctedpixel.y;
	bool propagation = CriticalConfiguration(u[newx][newy],u[newx+1][newy],u[newx][newy+1],u[newx+1][newy+1]);
	if(propagation){
	  Coords newCC;
	  newCC.x = newx;
	  newCC.y = newy; 
	  myqueue.push(newCC);
	}
      }
    }
  }
}

void TableToImage(unsigned ** im_in, image2d<value::int_u8> & im_out)
{
  mln_piter_(box2d) p(im_out.domain());
  for_all(p)
  {
    im_out(p) = im_in[p.row()][p.col()];
  }
}

void ImageToTable(image2d<value::int_u8> ima, unsigned ** u){
  point2d p_oo = ima.domain().pmin();
  mln_piter_(box2d) p(ima.domain());
  for_all(p)
  {
    u[p.row() - p_oo.row()][p.col() - p_oo.col()] = unsigned(ima(p));
  }
}


int main(int argc, const char * argv[] )
{

  if (argc != 2){
    std::cout <<"To use this program: \"./icip imagename.pgm\"\n";
  }
  else {

    //loading image
    std::cout <<"Loading Image ...\n";
    image2d<value::int_u8> ima;
    std::string filenameIN = argv[1];
    io::pgm::load(ima, filenameIN);

    //converting image to table
    std::cout <<"Converting Image To Table ...\n";
    unsigned sX = ima.nrows();
    unsigned sY = ima.ncols();
    std::cout <<"sX = " <<sX <<", sY = " <<sY <<'\n';
    unsigned ** u = alloue_image(ima.nrows(),ima.ncols());
    ImageToTable(ima,u);

    //statistics
    std::cout <<"Counting the critical configurations before the increasing process ...\n";
    std::cout <<nbCCs(u,sX,sY) <<" found\n";

    //applying increasing process
    std::cout <<"Applying the increasing process\n";
    increasingFIFO(u,sX,sY);

    //statistics
    std::cout <<nbCCs(u,sX,sY) <<" remaining critical configurations\n";

    //saving the result
    image2d<value::int_u8> imaWC(ima.domain());
    std::cout <<"Converting Table To Image ...\n";
    TableToImage(u, imaWC);
    unsigned found = filenameIN.find_last_of(".");
    std::string filenameOUT = filenameIN.substr(0,found);
    filenameOUT.append("_WC.pgm");
    std::cout<<"Saving new image file to: " <<filenameOUT <<'\n';
    io::pgm::save(imaWC,filenameOUT);
    std::cout<<"New image file saved to: " <<filenameOUT <<'\n';


  }
}




